#!/usr/bin/env python2
#
# installer.py - format partitions and install system
#
# (c) Copyright 2010 Michael Towers (larch42 at googlemail dot com)
#
# This file is part of the larch project.
#
#    larch is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    larch is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with larch; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#----------------------------------------------------------------------------
# 2010.10.31

import threading, os
import backend, fstab

# Perhaps there should be the possiblitiy of installing to an already mounted
# destination (passing the path) - that would allow at least modular support
# of essentially non-supported scenarios.

MP = '/tmp/larchin_mnt'     # Mount point for destination partition
execfile('/etc/larchin.conf')

class Installer:
    """This class manages the copying of the live system to the disk.
    It also handles peripheral requirements such as partition formatting
    and mounting, and the generation of an initramfs and initial /etc/fstab,
    bootloader installation, etc., using external modules for some of the
    steps.
    """
    def __init__(self):
        self.mainlabel = options.mainlabel if options.mainlabel else mainlabel
        if self.swapsize == 0:
            self.swaplabel = None
        else:
            self.swaplabel = options.swaplabel if options.swaplabel else swaplabel
        if self.homesize == 0:
            self.homelabel = None
        else:
            self.homelabel = options.homelabel if options.homelabel else homelabel


    def count_files(self):
        """Count the files to be copied during the installation
        (for progress reporting). This is run in a background thread.
        """
        ok, res = run_capture('find', 'bin', 'boot', 'etc', 'lib', 'home',
                'opt', 'root', 'sbin', 'srv', 'usr', 'var',
                '-printf', '1', cwd='/')
        if ok:
            self.nfiles = len(res)


    def run(self):
        """This function is the main controller for the sequence of processes.
        """
        ### Do partitioning
        if options.autopart:
            errormessage = partitions.autopart(self.swaplabel, options.swapsize,
                    self.mainlabel, options.mainsize, mainsize,
                    self.homelabel, options.homesize)
            if errormessage:
                return errormessage

        ### Do formatting
        if options.format:
            # Reformat the main partition (identified by its label),
            # and also the user partition (identified by its label) if
            # self.homesize != 0.
            errormessage = partitions.format(self.mainlabel)
            if (not errormessage) and (self.homelabel):
                errormessage = partitions.format(self.homelabel)
            if errormessage:
                return errormessage

        ### Mount the system
        if options.mount:
            errormessage = backend.mount(self.mainlabel, MP)
            if (not errormessage) and (self.homelabel):
                mp_home = os.path.join(MP, 'home')
                run_capture('mkdir', '-p', mp_home)
                errormessage = backend.mount(self.homelabel, mp_home)
            if errormessage:
                return errormessage

        ### Perform the actual installation
        if options.do_copy:
            self.nfiles = 0
            if options.progress:
                # Start thread to count the files to be copied, -> self.nfiles
                t = threading.Thread(target=self.count_files, args=())
                t.start()

            errormessage = self.copysystem():
            if errormessage:
                return errormessage

            io.out("#>" + _("Copy finished:") + "%6.1f GB" %
                    (float(self.installed) / 10**9))
            if self.nfiles:
                io.out(":p>")         # Clear percent progress

            io.out("#>" + _("Initial tweaks to installed system"))
            if not scripts.run("fix-system1", MP):
                return _("Initial installed-system tweaks failed")

            io.out("#>" + _("Removing live-system modifications"))
            if not scripts.run("fix-system2", MP,
                    "sshkeys" if options.ssh else ""):
                return _("Failure while removing live-system modifications")

        ### initramfs
        if options.initramfs or options.do_copy:
            io.out("#>" + _("Generating initramfs"))
            if not backend.run_mount_devprocsys("do-mkinitcpio", MP):
                return _("Problem building initramfs")

        ### /etc/fstab
        if options.fstab or options.do_copy:
            io.out("#>" + _("Generating /etc/fstab"))
#TODO:
            fst = fstab.generate(???)

            if not (fst and backend.writefile(fst, os.path.join(MP, '/etc/fstab'))):
                return _("Couldn't write /etc/fstab")

        ### extlinux
        if options.extlinux or options.do_copy:
            if not options.nobootloader:
#TODO:
# Perhaps a bootlines file from the profile might end up in /larch ?
# Otherwise just add a default?
# What about other systems, including 'toolkit'/'larch'?
# Or what about an interactive series of requests for titles for
# partitions that show up in blkid?
# Can use chain bootloader - but I think this can't use label or uuid?

        ### Master Boot Record
        if options.mbr:
#TODO:




    def copysystem(self):
        io.out("#>" + _("Copying system to selected partition(s)"))
        rootdir = ""
        self.installed = 0      # bytes
        mbi = 0                 # MB
        files = 0
        percent = 0
        fail = 0

        p = scripts.start_script("copy-system", MP)
        while True:
            line = p.stdout.readline()
            if not line:
                break

            if "!" not in line:
                if fail < 4:
                    fail += 1
                    debug("copy-system -> " + line)
                continue
            fname, size = line.split("!")
            frdir = fname.lstrip("/").split("/", 1)[0]
            if frdir != rootdir:
                rootdir = frdir
                io.out("#>" + " ... /%s ..." % frdir)
            self.installed += int(size)
            files += 1
            mb = (self.installed + 5*10**5) / 10**6
            if (options.mb > 0) and ((mb - mbi) >= options.mb):
                io.out(":m>%d" % mb)
                mbi = mb
            if self.nfiles:
                pc = files * 100 / self.nfiles
                if (pc > percent) and pc <= 100:
                    io.out(":p>%d" % pc)
                    percent = pc

        # The process has ended.
        return None if scripts.end_script(p) else _("Copying of system data failed")



if __name__ == "__main__":
    backend.start_translator()

    from optparse import OptionParser, OptionGroup
    parser = OptionParser(usage=_("usage: %prog [options]"))

    parser.add_option("-a", "--autopart", action="store_true",
            default="False", dest="autopart",
            help=_("Allocate partitions for the installation automatically"))
    parser.add_option("-f", "--format", action="store_true",
            default="False", dest="format",
            help=_("Format the partition(s)"))
    parser.add_option("-i", "--install", action="store_true",
            default="False", dest="install",
            help=_("Install the live system to the destination partition(s)"))
    parser.add_option('-I', '--only-install', action='store', type='string',
            default='', dest='installation_path',
            help=_("Install the live system to an already mounted destination"),

    parser.add_option("-g", "--only-mkinitcpio", action="store_true",
            default="False", dest="initramfs",
            help=_("Only generate the initramfs"))
    parser.add_option("-t", "--only-fstab", action="store_true",
            default="False", dest="fstab",
            help=_("Only generate /etc/fstab"))
    parser.add_option("-b", "--only-extlinux", action="store_true",
            default="False", dest="extlinux",
            help=_("Only install the extlinux bootloader"))
    parser.add_option("-B", "--nobootloader", action="store_true",
            default="False", dest="nobootloader",
            help=_("Don't install the bootloader"))
    parser.add_option("-r", "--mbr", action="store_true",
            default="False", dest="mbr",
            help=_("Install a Master Boot Record to the main drive and mark"
            " the installation partition as bootable"))


    parser.add_option('-S', '--swap_label', action='store', type='string',
            default='', dest='swaplabel', help=_("Label for the swap partition"),
    parser.add_option('-s', '--swap_size', action='store', type='float',
            default='-1', dest='swapsize',
            help=_("Size (GB - float) for the swap partition; 0 => no swap partition"),
    parser.add_option('-M', '--main_label', action='store', type='string',
            default='', dest='mainlabel', help=_("Label for the main ('/') partition"),
    parser.add_option('-m', '--main_size', action='store', type='float',
            default='-1', dest='mainsize',
            help=_("Size (GB - float) for the main partition"),
    parser.add_option('-U', '--home_label', action='store', type='string',
            default='', dest='homelabel', help=_("Label for the home partition"),
    parser.add_option('-u', '--home_size', action='store', type='float',
            default='-1', dest='homesize',
            help=_("Size (GB - float) for the home partition; 0 => no home partition"),

    parser.add_option("-P", "--sizemb", action="store", dest="mb",
            type="int", default=0,
            help=_("Output installed MB (progress) with step of this size"))
    parser.add_option("-p", "--progress", action="store_true", dest="progress",
            default=False, help=_("Output estimated progress as percent"))
    parser.add_option("-q", "--quiet", action="store_true", dest="quiet",
            default=False, help=_("Suppress output messages, except errors"))
    parser.add_option("-k", "--keepsshhost", action="store_true", dest="ssh",
            default=False, help=_("Retain ssh_host keys (generally not a good idea)"))
    (options, args) = parser.parse_args()

    backend.init(options.quiet)

    # Add individual step flags
    options.mount = False
    options.do_copy = True
    if options.initramfs or options.fstab or options.extlinux:
        options.format = False
        options.autopart = False
        options.mount = True
        options.do_copy = False

    if options.installation_path:
        options.format = False
        options.autopart = False
        options.install = False
        options.mount = False
        MP = options.installation_path
    elif options.install:
        options.mount = True
    else:
        options.do_copy = False

    if options.autopart:
        options.format = False

    installer = Installer()

    sys_quit(installer.run())
